/*
**	BPA connection idle timeout check, if the idle time reached, will disconnect the BPA
**
*/ 
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include "value_parser_file.h"

#define WAN_UPTIME "/var/wan_uptime"
#define ETH_WAN         "eth0"
#define PPP0_WAN	"ppp0"
#define DOD_IP "10.64.64.64"

#define WAN_GetStatusOK             0
#define WAN_GetStatusErr            -1

typedef struct{
    unsigned char mac[6];
    short flag;
    struct in_addr ip;
    struct in_addr mask;
    struct in_addr gw;    
}IFACE_STATUS;
IFACE_STATUS if_status;

struct sysinfo {
	long uptime;			/* Seconds since boot */
	unsigned long loads[3];		/* 1, 5, and 15 minute load averages */
	unsigned long totalram;		/* Total usable main memory size */
	unsigned long freeram;		/* Available memory size */
	unsigned long sharedram;	/* Amount of shared memory */
	unsigned long bufferram;	/* Memory used by buffers */
	unsigned long totalswap;	/* Total swap space size */
	unsigned long freeswap;		/* swap space still available */
	unsigned short procs;		/* Number of current processes */
	unsigned short pad;			/* Padding needed for m68k */
	unsigned long totalhigh;	/* Total high memory size */
	unsigned long freehigh;		/* Available high memory size */
	unsigned int mem_unit;		/* Memory unit size in bytes */
	char _f[20-2*sizeof(long)-sizeof(int)];	/* Padding: libc5 uses this.. */
};
extern int sysinfo (struct sysinfo* info);

#if 1
#include <stdarg.h>
void mBUG(char *format,...){
 va_list args;
    FILE *fp;

    fp = fopen("/var/cgitest","a+");
    if(!fp){
        return;
    }
    va_start(args,format);
    vfprintf(fp,format,args);
    va_end(args);
    fprintf(fp,"\n");
    fflush(fp);
    fclose(fp);
    system("/bin/chmod 777 /var/cgitest");

}
#else 
#define mBUG(format,...) 
#endif

/*
**  Main processing loop.  Logs in, handles heartbeats, and logs out on request
**
**  Returns - 0 - We are no longer connected, and should not retry.
**            1 - We are no longer connnected and should retry.
*/
int main(int argc,char **argv)
{
    struct sysinfo info;
	unsigned long last_pkttime, last_pktnum, bpa_idletime = 0;
	FILE *fp;
	char tmp[16];
	char *temp;
	int bpa_con_stat;
	unsigned long counters[6];

//    mBUG("go to the main of bpaidle");	

    last_pkttime = last_pktnum = 0;
	temp = value_parser_file("wan_mode");
//	mBUG("in the main of bpaidle, wan_mode =%s", temp);	
	if (strcmp(temp,"bpa") != 0)            //the wan connection type is not bpa
	    return 0;

    temp = value_parser_file("bpa_idle");
    bpa_idletime = atoi(temp);
    
	
    while(1)
    {
        sleep(5);
        sysinfo(&info);
        bpa_con_stat = bpa_con_status();
        
//        mBUG("in the main of bpaidle, bpa_con_stat =%d, (int)info.uptime =%d, bpa_idletime=%d", bpa_con_stat, (int)info.uptime, bpa_idletime);
        
        if(0 == bpa_idletime)           //the bpa idle time is 0, will keep alive
        {
            if(0 == bpa_con_stat)
            {
//                mBUG("in the main of bpaidle, the idle time is 0, bpa is not connected, will keep the connection alive.");
//        	    system("/bin/killall udhcpc");
        	    sleep(2);
                system("/bin/killall -SIGUSR1 udhcpc");
	            sleep(3);
	            system("/bin/bpa start");
            }
        }
        else
        {
            if(1 == bpa_con_stat)       //the bpa connection is connected, will check the timeout
            {
            	rtl8651_getIntfCounters(0x1, &counters);
            	if(counters[3] > last_pktnum)
            	    last_pkttime = info.uptime;
//                mBUG("in the main of bpaidle, last_pktnum =%lu, counters[3] =%lu, last_pkttime =%lu", last_pktnum, counters[3], last_pkttime);
                last_pktnum = counters[3];
                            	
                if((last_pkttime > 0) && ((info.uptime - last_pkttime) > bpa_idletime*60))
                {
                    mBUG("in the main of bpaidle, the idle time is %d*60s, bpa is connected, (int)info.uptime =%d, last_pkttime =%d, will disconnect the BPA connection.", bpa_idletime, (int)info.uptime, last_pkttime);
                    system("/bin/killall -SIGUSR2 udhcpc");
                    sleep(3);
                    system("/bin/bpa stop");
	            }
            }
        }
    }

	return 1;
}

/* check the wan bpa connection status 
    1 --- bpa connected
    0 --- no connection
*/
static int bpa_con_status(void)
{
	char buf[20];
	
	GetBpaConnectionStatus(&if_status);
	strcpy(buf,inet_ntoa(if_status.ip));
	
    if((access(WAN_UPTIME, F_OK) == 0) && (strcmp(buf,"0.0.0.0") != 0))
    {
        return 1;
	}
	else
	{
	    return 0;
	}
}

static int GetBpaConnectionStatus(IFACE_STATUS *pStatus)
{
    char if_name[6];
    char mac[6];

    /* Try to get connection Type. */
    char *pt=value_parser_file("wan_mode");

    memset(pStatus,0,sizeof(IFACE_STATUS));

	/* For I can not get ppp iface's MAC ,get ethx MAC first.*/
	GetIfaceStatus(ETH_WAN,pStatus);
	memcpy(mac,pStatus->mac,6);
	
	if(!strcmp(pt,"dhcpc") || !strcmp(pt, "ip") || !strcmp(pt, "bpa"))
		strcpy(if_name, ETH_WAN);
	else if((strcmp(pt,"pppoe")==0) || (strcmp(pt,"pptp")==0))
		strcpy(if_name,"ppp0");
	else
		return WAN_GetStatusErr;

    GetIfaceStatus(if_name,pStatus);
    /* Then copy MAC back. */
    memcpy(pStatus->mac,mac,6);

    return WAN_GetStatusOK;
}

/********************************************************************************
 *  GetIfaceStatus()        Get interface status and put the data to *pStatus   *
 *                                                                              *
 *  0                   ==>     Successed                                       *
 *  other               ==>     Fail                                            *
 *                                                                              *
 ********************************************************************************/
int GetIfaceStatus(char *if_name, IFACE_STATUS *pStatus){
    struct ifreq ifr;
    int sockd;
    struct sockaddr_in *mysockaddr;

    memset(pStatus,0,sizeof(IFACE_STATUS));
    sockd = socket(AF_INET, SOCK_STREAM, 0);
    if (sockd == -1) { /* Open socket error. */
		return WAN_GetStatusErr;
	}
	/* Init iface name. */
	strcpy(ifr.ifr_name, if_name);

    /* Try to get HW address. */
    if(ioctl(sockd, SIOCGIFHWADDR, &ifr) < 0) {
    		close(sockd);
		return WAN_GetStatusErr;
	}
	memcpy(pStatus->mac,ifr.ifr_hwaddr.sa_data,6);

    /* Try to get IP address. */
    if(ioctl(sockd, SIOCGIFADDR, &ifr) < 0) { ///This is not fitable for pppoe,because 10.64.64.64. so add following, gavin,2004-11-13
    		close(sockd);
		return WAN_GetStatusErr;
	}
	mysockaddr = (struct sockaddr_in *)&(ifr.ifr_addr);
	pStatus->ip = mysockaddr->sin_addr;
       ///gavin added the following ^M 
       if(!strcmp(if_name,PPP0_WAN) && !strcmp(inet_ntoa(pStatus->ip),DOD_IP)) 
         {
                close(sockd);
                return WAN_GetStatusErr;
        }
       /* Try to get netmask. */
    if(ioctl(sockd, SIOCGIFNETMASK, &ifr) < 0) {
    		close(sockd);
		return WAN_GetStatusErr;
	}
	mysockaddr = (struct sockaddr_in *)&(ifr.ifr_netmask);
	pStatus->mask = mysockaddr->sin_addr;

#if 0
    /* Try to get gw. */
    if(ioctl(sockd, SIOCGIFDSTADDR, &ifr) < 0) {
    		close(sockd);
		return WAN_GetStatusErr;
	}
	mysockaddr = (struct sockaddr_in *)&(ifr.ifr_addr);
	pStatus->gw = mysockaddr->sin_addr;
#elif0
{
    FILE *fp;
    char sys_cmd[128];
    #define ROUTE_RESULT    "/var/routeResult"

    sprintf(sys_cmd,"/sbin/route -n | grep \"^0.0.0.0\" > %s 2>&1",ROUTE_RESULT);
    system(sys_cmd);

    fp = fopen(ROUTE_RESULT,"r");
    if(fp){

        char line[128];
        int i;
        while(fgets(line,sizeof(line),fp) != NULL){
            for(i=16;line[i] != ' ';i++);
            line[i] = '\0';
            inet_aton(&line[16],&pStatus->gw);
        }
        fclose(fp);
    }
    remove(ROUTE_RESULT);
}
#else
{
	FILE *fp;
	fp=fopen("/proc/net/route", "r");
	if(fp){
		char buf[128];
		while(fgets(buf, sizeof(buf), fp)){
			if(!strstr(buf, if_name))
				continue;
			if(sscanf(buf, "%*s\t00000000\t%lx",(long int *)&pStatus->gw.s_addr))
				break;
		}
		fclose(fp);
	}
}
#endif

    /* Try to get flag. */
    if(ioctl(sockd, SIOCGIFFLAGS, &ifr) < 0) {
    		close(sockd);
		return WAN_GetStatusErr;
	}
	pStatus->flag = ifr.ifr_flags;
///debug1("flag=%x", pStatus->flag);	
    close(sockd);

    return WAN_GetStatusOK;
}