/* 2004.05.21  fix memory leak */
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <string.h>
#include <sys/sysinfo.h>
#include "upnpd.h"
#include "tool.h"

/* Ron */
#include "dll.h"
extern int IGDDeviceStateTableInit(char*,char*);
extern int IGDdelAllPortForwarding(void);
extern int substr(char *infile, char *outfile, char *str_from, char *str_to);

struct GateEnv gate;

int custom_ttl;
int advertisement_expire=5;

extern int IGDDeviceCallbackEventHandler(Upnp_EventType, void*, void*);



int main(int argc, char** argv)
{

    int ret=1;
    int port=0;
    char *desc_doc_name=NULL, *web_dir_path=NULL;
    int sig;
    char fullurl[200];
    char desc_doc_url[200];
    struct sysinfo info;
    sigset_t sigs_to_catch;
    
    bzero(&gate,sizeof(struct GateEnv));
    
    if (argc != 5){
		printf("Usage: upnpd <ext ifname> <int ifname> <adv time> <ttl>\n");
		printf("Example: upnpd ppp0 eth0 30 4\n");
		printf("Example: upnpd eth1 eth0 10 1\n");
		exit(0);
    }
        
    gate.int_ifname=argv[2];
    gate.ext_ifname=argv[1];  

    strcpy(gate.InternalIPAddress,getIPAddress(gate.int_ifname));
//    strcpy(gate.ExternalIPAddress,getIPAddress(gate.ext_ifname)?:"0.0.0.0");
//    printf("ExternalIPAddress=%s\n",gate.ExternalIPAddress);
    getConnectionStatus();
    getLayer1MaxBitRate();    
  
    sysinfo(&info);
    gate.startup_time = info.uptime;
//  gate.startup_time = time(NULL);       
    custom_ttl=atoi(argv[4]);
  //gate.advr_expire=(atoi(argv[3])*60)*2+60;
    gate.advr_expire=(atoi(argv[3])*60)*2;
    
    
    desc_doc_name=DEF_DOCUMENT;  
    web_dir_path=DEF_DIR_PATH;  

    substr("/var/upnp/gateway.mod", "/var/upnp/gateway.xml", "IPADDR",gate.InternalIPAddress);
    //substr("/tmp/gateway.mod", "/tmp/gateway.xml", "IPADDR",gate.InternalIPAddress);
    if ((ret = UpnpInit(gate.InternalIPAddress, port)) != UPNP_E_SUCCESS) {
	printf("Error with UpnpInit -- %d\n", ret);
	UpnpFinish();
	exit(1);
    }
    printf("UPnP Initialized...\n");
   
    if (port==0)
        port = UpnpGetServerPort();

     sprintf(fullurl, "http://%s:%d/%s", gate.InternalIPAddress , port, desc_doc_name);
     sprintf(desc_doc_url, "http://%s:%d", gate.InternalIPAddress , port);
    
    printf("Intialized UPnP with fullurl=%s\n", fullurl);
    printf("ipaddress=%s port=%d\n", gate.InternalIPAddress, port);
    printf("web_dir_path=%s\n", web_dir_path);
    printf("desc_doc_url=%s\n", desc_doc_url);

	printf("Specifying the webserver root directory -- %s\n", web_dir_path);
    if ((ret = UpnpSetWebServerRootDir(web_dir_path)) != UPNP_E_SUCCESS) {
	printf("Error specifying webserver root directory -- %s: %d\n", web_dir_path, ret);
	UpnpFinish();
	exit(1);
    }

    printf("Registering the RootDevice\n");
    if ((ret = UpnpRegisterRootDevice(fullurl, IGDDeviceCallbackEventHandler, &gate.device_handle, &gate.device_handle)) != UPNP_E_SUCCESS) {
	printf("Error registering the rootdevice : %d\n", ret);
	UpnpFinish();
	exit(1);
    } else {
	printf("RootDevice Registered\n");
     
    //Insert the first default PortMaping in dll;
    AddDefaultNode();
     
	printf("Initializing State Table\n");
	IGDDeviceStateTableInit(desc_doc_url,desc_doc_name);
	if ((ret = UpnpSendAdvertisement(gate.device_handle, gate.advr_expire)) != UPNP_E_SUCCESS) {
	    printf("Error sending advertisements : %d\n", ret);
	    UpnpFinish();
	    exit(1);
	}
	printf("UPNP Advertisements Sent\n");
    }

    /* Catch Ctrl-C and properly shutdown */
    sigemptyset(&sigs_to_catch);
    sigaddset(&sigs_to_catch, SIGINT);
    sigaddset(&sigs_to_catch, SIGTERM);
    sigwait(&sigs_to_catch, &sig);
    printf("Shutting down on signal %d...\n", sig);
    UpnpUnRegisterRootDevice(gate.device_handle);
    UpnpFinish();

    IGDdelAllPortForwarding();
    /* remove the upnptab file if exist */
//    if(access("/var/upnptab",F_OK)==0)
//    	remove("/var/upnptab");
    exit(0);
}


