#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/utsname.h>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>
#include <sys/timex.h>
#include "libntp.h"

#define JAN_1970        0x83aa7e80      /* 2208988800 1970 - 1900 in seconds */
#define NTPFRAC(x) ( 4294*(x) + ( (1981*(x))>>11 ) )
#define USEC(x) ( ( (x) >> 12 ) - 759 * ( ( ( (x) >> 10 ) + 32768 ) >> 16 ) )
#define sec2u(x) ( (x) * 15.2587890625 )

#define dprintf printf

extern int contemplate_data(unsigned int absolute, double skew, double errorbar, int freq);
	
int setup_ntp_socket(unsigned int interface, short port)
{
	struct sockaddr_in sa_recv;
	int sockfd=-1;
	
	if ((sockfd=socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP))==-1)	
		return -1;
	
	bzero((char *) &sa_recv, sizeof(sa_recv));
	sa_recv.sin_family=AF_INET;
	sa_recv.sin_addr.s_addr=htonl(interface);
	sa_recv.sin_port=htons(port);
	
	
	if(bind(sockfd,(struct sockaddr *) &sa_recv,sizeof(sa_recv)) == -1) {
		perror("bind");
		return -1;
	}
	return sockfd;
}


int send_request(int sockfd,char *host, short port)
{
	u_int32_t data[12];
	struct timeval now;
	struct hostent *ntpserver;
	struct sockaddr_in sa_dest;
	
	bzero((char *) &sa_dest, sizeof(sa_dest));
	sa_dest.sin_family=AF_INET;
	
	ntpserver=gethostbyname(host);
	
	if (ntpserver == NULL) {
		return DNS_ERR;
	}
	if (ntpserver->h_length != 4) {
		return DNS_ERR;
	}
	memcpy(&(sa_dest.sin_addr.s_addr),ntpserver->h_addr_list[0],4);

	sa_dest.sin_port=htons(port);
	
	if (connect(sockfd,(struct sockaddr *)&sa_dest,sizeof(sa_dest))==-1){
		perror("connect");
		return NTP_ERR;
	}
	
	bzero(data,sizeof(data));
#define LI 0
#define VN 3
#define MODE 3
#define STRATUM 0
#define POLL 4 
#define PREC -6

	data[0] = htonl (
		( LI << 30 ) | ( VN << 27 ) | ( MODE << 24 ) |
		( STRATUM << 16) | ( POLL << 8 ) | ( PREC & 0xff ) );
	data[1] = htonl(1<<16);  /* Root Delay (seconds) */
	data[2] = htonl(1<<16);  /* Root Dispersion (seconds) */
	gettimeofday(&now,NULL);
	data[10] = htonl(now.tv_sec + JAN_1970); /* Transmit Timestamp coarse */
	data[11] = htonl(NTPFRAC(now.tv_usec));  /* Transmit Timestamp fine   */
	send(sockfd,data,48,0);
	return NTP_OK;	
}

int recv_response(int sockfd,char *buf,unsigned int timeout,char *ntpserver)
{
	int len=0,packet_len=0;
	char data[1500];
	struct sockaddr sa_recv;
	struct sockaddr_in *sa_in;
	fd_set rfds;
	struct timeval tv;
	int retval;
	FD_ZERO(&rfds);
	FD_SET(sockfd, &rfds);
	tv.tv_sec = timeout;
	tv.tv_usec = 0;
	
	retval = select(sockfd+1, &rfds, NULL, NULL, &tv);
	
	if (retval<=0)
		return NTP_TIMEOUT;
	
	len=sizeof(sa_recv);
	
	if((packet_len=recvfrom(sockfd,data,sizeof(data),0,&sa_recv,&len))<0)
		return NTP_ERR;
	
	if (sa_recv.sa_family==AF_INET && ntpserver!=NULL) {
		sa_in=(struct sockaddr_in *)&sa_recv;
		strcpy(ntpserver,inet_ntoa(sa_in->sin_addr));
	}
	memcpy(buf,data,packet_len);
	return NTP_OK;	
}

int get_current_freq(void)
{
	struct timex txc;
	txc.modes=0;
	if (__adjtimex(&txc) < 0) {
		return -1;
	}
	return txc.freq;
}

int set_freq(int new_freq)
{
	struct timex txc;
	txc.modes = ADJ_FREQUENCY;
	txc.freq = new_freq;
	if (__adjtimex(&txc) < 0) {
		return -1;
	}
	return txc.freq;
}

struct ntptime {
	unsigned int coarse;
	unsigned int fine;
};

double ntpdiff( struct ntptime *start, struct ntptime *stop)
{
	int a;
	unsigned int b;
	a = stop->coarse - start->coarse;
	if (stop->fine >= start->fine) {
		b = stop->fine - start->fine;
	} else {
		b = start->fine - stop->fine;
		b = ~b;
		a -= 1;
	}
	
	return a*1.e6 + b * (1.e6/4294967296.0);
}

int handle_ntp_reply(char *data)
{
	/* straight out of RFC-1305 Appendix A */
	int mode, stratum,disp;
	struct ntptime reftime, orgtime, rectime, xmttime,arrival;
	struct timeval tv_set,udp_arrival;
	double etime,stime,skew1,skew2;
	int new_freq,freq;

	mode   = ntohl(((unsigned int *)data)[0]) >> 24 & 0x07;
	stratum = ntohl(((unsigned int *)data)[0]) >> 16 & 0xff;
	disp    = ntohl(((unsigned int *)data)[2]);
	reftime.coarse = ntohl(((unsigned int *)data)[4]);
	reftime.fine   = ntohl(((unsigned int *)data)[5]);
	orgtime.coarse = ntohl(((unsigned int *)data)[6]);
	orgtime.fine   = ntohl(((unsigned int *)data)[7]);
	rectime.coarse = ntohl(((unsigned int *)data)[8]);
	rectime.fine   = ntohl(((unsigned int *)data)[9]);
	xmttime.coarse = ntohl(((unsigned int *)data)[10]);
	xmttime.fine   = ntohl(((unsigned int *)data)[11]);

	/* it would be even better to subtract half the slop */
	tv_set.tv_sec  = xmttime.coarse - JAN_1970;
	/* divide fine by 4294.967296 */
	tv_set.tv_usec = USEC(xmttime.fine);
	

	if(mode==4 && stratum!=0 && xmttime.coarse!=0 ){
		if (settimeofday(&tv_set,NULL)<0) {
			printf("settimeofday failed!\n");
			return NTP_ERR;
		}	
	}else{
		dprintf("mode==%d\n stratum==%d\n gettime %lu.%.6lu\n",
			       	mode,stratum,tv_set.tv_sec, tv_set.tv_usec);
		if(stratum==0)
			return NTP_KOD;
		
		return NTP_ERR;
	}
	
	/* System Clock Accuracy */	
	gettimeofday(&udp_arrival,NULL);
	arrival.coarse = udp_arrival.tv_sec + JAN_1970;
	arrival.fine   = NTPFRAC(udp_arrival.tv_usec);

	etime=ntpdiff(&orgtime,&arrival);
	stime=ntpdiff(&rectime,&xmttime);
	skew1=ntpdiff(&orgtime,&rectime);
	skew2=ntpdiff(&xmttime,&arrival);
	freq=get_current_freq();
	
	new_freq = contemplate_data(arrival.coarse, (skew1-skew2)/2,
		etime+sec2u(disp), freq);
	if (new_freq != freq) set_freq(new_freq);

	//dprintf("%d %d\n",new_freq,freq);
	return NTP_OK;
}


